/* ********************************************************************
    Licensed to Jasig under one or more contributor license
    agreements. See the NOTICE file distributed with this work
    for additional information regarding copyright ownership.
    Jasig licenses this file to you under the Apache License,
    Version 2.0 (the "License"); you may not use this file
    except in compliance with the License. You may obtain a
    copy of the License at:
        
    http://www.apache.org/licenses/LICENSE-2.0
        
    Unless required by applicable law or agreed to in writing,
    software distributed under the License is distributed on
    an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
    KIND, either express or implied. See the License for the
    specific language governing permissions and limitations
    under the License.
*/
package org.bedework.tools.cmdutil;

import org.bedework.calfacade.filter.SfpTokenizer;
import org.bedework.calsvci.CalSvcFactoryDefault;
import org.bedework.calsvci.CalSvcI;
import org.bedework.calsvci.CalSvcIPars;

import edu.rpi.sss.util.Args;

import org.apache.log4j.Logger;

import java.io.FileInputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.StreamTokenizer;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

/** Utility to execute various operations specified by commands.
 *
 * <p>Takes one or more lines of input specifying the action. Lines starting
 * with "#" are treated as comments.
 *
 * Non-comment lines are:
 * <pre>
 * create category val
 * </pre>
 *
 * <p>Creates a category. val must be quoted if it contains non-work characters.
 *
 * <pre>
 * create collection (folder| calendar) parent-path name
 *            [filter=qfexpr] (category=cat)* [owner]
 * create collection alias parent-path name alias-target
 *            [filter=qfexpr] (category=cat)* [owner]
 * </pre>
 *
 * <p>Creates a folder or calendar collection under the given parent with the
 * given name and optional owner.
 *
 * <p>qfexpr is a quoted filter expression.
 *
 * <pre>
 * list categories val*
 * </pre>
 *
 * <p>List categories. If no val is supplied all categories are listed. Any
 * supplied values are matched against the categories. Any case insignificant
 * match results in the category being listed.
 *
 * <pre>
 *   move events [addcats] from to category*
 * </pre>
 *
 * <p>where from and to are possibly quoted paths, e.g. /public/Arts/Dance
 * and category* is 0 or more possibly quoted categories to set on the entities
 * events as they are moved.
 *
 * <p>If addcats is specified the events will have categories set derived from
 * the from path. e.g. if from="/public/Arts/Dance" then the event will have
 * categories "Arts" and "Dance" (which must exist)
 *
 * <pre>
 *   stop
 * </pre>
 *
 * <p>Terminate the utility
 *
 * @author douglm
 */
public class CmdUtil {
  private boolean echo = false;

  private transient Logger log;

  private String adminUserAccount;

  private String infileName;

  static class ProcessState {
    boolean debug;

    CalSvcI svci;

    SfpTokenizer tokenizer;

    Collection<String> errors;
  }

  private ProcessState pstate;

  private Map<String, CmdUtilHelper> processors = new HashMap<String, CmdUtilHelper>();

  CmdUtil()  throws Throwable {
    pstate = new ProcessState();

    processors.put("create", new ProcessCreate(pstate));
    processors.put("delete", new ProcessDelete(pstate));
    processors.put("list", new ProcessList(pstate));
    processors.put("move", new ProcessMove(pstate));
  }

  void process() throws Throwable {
    pstate.debug = getLog().isDebugEnabled();

    getSvci();
    InputStream is;

    if (infileName != null) {
      is = new FileInputStream(infileName.trim());
      echo = true;
    } else {
      is = System.in;
    }

    LineNumberReader lis = new LineNumberReader(new InputStreamReader(is));

    for (;;) {
      pstate.errors = new ArrayList<String>();

      String ln = lis.readLine();

      if (ln == null) {
        break;
      }

      if (echo) {
        info(ln);
      }

      if (ln.startsWith("#")) {
        continue;
      }

      int pst = processLine(ln);

      if (pst == eof) {
        break;
      }

      if (pst == err) {
        error("Failed command: " + ln);
      }

      for (String s: pstate.errors) {
        error(s);
      }
    }
  }

  private final static int eof = 0;

  private final static int ok = 1;

  private final static int err = 2;

  private int processLine(final String ln) throws Throwable {
    pstate.tokenizer = new SfpTokenizer(new StringReader(ln));

    String wd = word();

    if (wd == null) {
      if (pstate.tokenizer.atEof()) {
        // Blank line
        return ok;
      }

      return err;
    }

    if ("stop".equals(wd)) {
      return eof;
    }

    CmdUtilHelper cuh = processors.get(wd);
    if (cuh == null) {
      return err;
    }

    if (cuh.process()) {
      return ok;
    }

    return err;
  }

  private String word() throws Throwable {
    if (pstate.tokenizer.testToken(StreamTokenizer.TT_WORD)) {
      return pstate.tokenizer.sval;
    }

    return null;
  }

  private void getSvci() throws Throwable {
    CalSvcIPars pars = new CalSvcIPars(adminUserAccount,
                                       null,   // calsuite
                                       true,   // publicAdmin
                                       true,   // superUser,
                                       true,   // service
                                       true,  // adminCanEditAllPublicCategories
                                       true,  // adminCanEditAllPublicLocations
                                       true,  // adminCanEditAllPublicSponsors
                                       false,    // sessionless
                                       null); // dbpars
    pstate.svci = new CalSvcFactoryDefault().getSvc(pars);
  }

  boolean processArgs(final Args args) throws Throwable {
    if (args == null) {
      return true;
    }

    while (args.more()) {
      if (args.ifMatch("")) {
        continue;
      }

      if (args.ifMatch("-debug")) {
        pstate.debug = true;
      } else if (args.ifMatch("-ndebug")) {
        pstate.debug = false;
      } else if (args.ifMatch("-f")) {
        infileName = args.next();
      } else if (args.ifMatch("-user")) {
        adminUserAccount = args.next();
      } else {
        error("Illegal argument: " + args.current());
        usage();
        return false;
      }
    }

    return true;
  }

  void usage() {
    System.out.println("Usage:");
    System.out.println("args   -debug");
    System.out.println("       -ndebug");
    System.out.println("       -f <filename>");
    System.out.println("            specify file containing commands");
    System.out.println("");
  }

  protected Logger getLog() {
    if (log == null) {
      log = Logger.getLogger(this.getClass());
    }

    return log;
  }

  protected void info(final String msg) {
    getLog().info(msg);
  }

  protected void error(final String msg) {
    getLog().error(msg);
  }

  protected void trace(final String msg) {
    getLog().debug(msg);
  }

  /** Main
   *
   * @param args
   */
  public static void main(final String[] args) {
    CmdUtil mc = null;

    try {
      mc = new CmdUtil();

      if (!mc.processArgs(new Args(args))) {
        return;
      }

      mc.process();
    } catch (Throwable t) {
      t.printStackTrace();
    }
  }
}
