/* ********************************************************************
    Licensed to Jasig under one or more contributor license
    agreements. See the NOTICE file distributed with this work
    for additional information regarding copyright ownership.
    Jasig licenses this file to you under the Apache License,
    Version 2.0 (the "License"); you may not use this file
    except in compliance with the License. You may obtain a
    copy of the License at:
        
    http://www.apache.org/licenses/LICENSE-2.0
        
    Unless required by applicable law or agreed to in writing,
    software distributed under the License is distributed on
    an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
    KIND, either express or implied. See the License for the
    specific language governing permissions and limitations
    under the License.
*/
package org.bedework.tools.cmdutil;

import org.bedework.calfacade.BwCalendar;
import org.bedework.calfacade.BwCategory;
import org.bedework.calfacade.BwString;
import org.bedework.calfacade.RecurringRetrievalMode;
import org.bedework.calfacade.RecurringRetrievalMode.Rmode;
import org.bedework.calfacade.exc.CalFacadeException;
import org.bedework.calfacade.svc.EventInfo;
import org.bedework.calsvci.CalSvcI;
import org.bedework.tools.cmdutil.CmdUtil.ProcessState;

import org.apache.log4j.Logger;

import java.io.StreamTokenizer;

/** Handle processing for CmdUtil
 *
 * @author douglm
 *
 */
public abstract class CmdUtilHelper {
  private transient Logger log;

  protected ProcessState pstate;

  CmdUtilHelper(final ProcessState pstate) {
    this.pstate = pstate;
  }

  abstract boolean process() throws Throwable;

  protected CalSvcI getSvci() {
    return pstate.svci;
  }

  protected void open() throws Throwable {
    getSvci().open();
    getSvci().beginTransaction();
  }

  protected void close() throws Throwable {
    getSvci().endTransaction();
    getSvci().close();
  }

  /* expect a possibly quoted path next
   */
  protected BwCalendar getCal() throws Throwable {
    String path = wordOrQuotedVal();

    if (path == null) {
      error("Expected a path");
      return null;
    }

    BwCalendar cal = getSvci().getCalendarsHandler().get(path);

    if (cal == null) {
      error("Unable to access calendar " + path);
    }

    return cal;
  }

  /* Get an event given path and name
   */
  protected EventInfo getEvent(final String path,
                               final String name) throws Throwable {
    if (path == null) {
      error("Expected a path");
      return null;
    }

    RecurringRetrievalMode rrm = new RecurringRetrievalMode(Rmode.overrides);

    EventInfo ent = getSvci().getEventsHandler().get(path, name, rrm);

    if (ent == null) {
      error("Unable to fetch event " + path + "/" + name);
    }

    return ent;
  }

  protected BwCategory getCat(final String ownerHref)  throws Throwable {
    String catVal = wordOrQuotedVal();

    if (catVal == null) {
      error("Expected a category");
      return null;
    }

    return getCat(ownerHref, catVal);
  }

  protected BwCategory getCat(final String ownerHref,
                              final String catVal)  throws Throwable {
    BwCategory cat = getSvci().getCategoriesHandler().find(new BwString(null, catVal),
                                                           ownerHref);

    if (cat == null) {
      error("Unable to access category " + catVal + " for owner " + ownerHref);
    }

    return cat;
  }

  protected int nextToken() throws CalFacadeException {
    int tkn = pstate.tokenizer.next();

    if (!pstate.debug) {
      return tkn;
    }

    if (tkn == StreamTokenizer.TT_WORD) {
      trace("tkn = word: " + pstate.tokenizer.sval);
    } else if (tkn > 0) {
      trace("tkn = " + (char)tkn);
    } else {
      trace("tkn = " + tkn);
    }

    return tkn;
  }

  protected boolean cmdEnd() throws Throwable {
    return pstate.tokenizer.testToken(StreamTokenizer.TT_EOL) ||
           pstate.tokenizer.testToken(StreamTokenizer.TT_EOF);
  }

  protected String word() throws Throwable {
    if (pstate.tokenizer.testToken(StreamTokenizer.TT_WORD)) {
      return pstate.tokenizer.sval;
    }

    return null;
  }

  protected String wordOrQuotedVal() throws Throwable {
    String wd = word();
    if (wd != null) {
      return wd;
    }

    return quotedVal();
  }

  protected String quotedVal() throws Throwable {
    if (!pstate.tokenizer.testString()) {
      return null;
    }

    return pstate.tokenizer.sval;
  }

  protected Logger getLog() {
    if (log == null) {
      log = Logger.getLogger(this.getClass());
    }

    return log;
  }

  protected void info(final String msg) {
    getLog().info(msg);
  }

  protected void error(final String msg) {
    getLog().error(msg);
  }

  protected void trace(final String msg) {
    getLog().debug(msg);
  }
}
