/* ********************************************************************
    Licensed to Jasig under one or more contributor license
    agreements. See the NOTICE file distributed with this work
    for additional information regarding copyright ownership.
    Jasig licenses this file to you under the Apache License,
    Version 2.0 (the "License"); you may not use this file
    except in compliance with the License. You may obtain a
    copy of the License at:
        
    http://www.apache.org/licenses/LICENSE-2.0
        
    Unless required by applicable law or agreed to in writing,
    software distributed under the License is distributed on
    an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
    KIND, either express or implied. See the License for the
    specific language governing permissions and limitations
    under the License.
*/
package org.bedework.tools.cmdutil;

import org.bedework.calfacade.BwCalendar;
import org.bedework.calfacade.BwCategory;
import org.bedework.calfacade.BwLocation;
import org.bedework.calfacade.BwString;
import org.bedework.calfacade.BwUser;
import org.bedework.calfacade.exc.CalFacadeException;
import org.bedework.tools.cmdutil.CmdUtil.ProcessState;

import java.util.ArrayList;
import java.util.List;

/**
 * @author douglm
 *
 */
public class ProcessCreate extends CmdUtilHelper {
  ProcessCreate(final ProcessState pstate) {
    super(pstate);
  }

  @Override
  boolean process() throws Throwable {
    String wd = word();

    if (wd == null) {
      return false;
    }

    if ("collection".equals(wd)) {
      return createCollection();
    }

    if ("category".equals(wd)) {
      return processCreateCategory(wordOrQuotedVal());
    }

    if ("location".equals(wd)) {
      return createLocation();
    }

    return false;
  }

  private boolean processCreateCategory(final String catVal) throws Throwable {
    try {
      if (catVal == null) {
        error("Expected a category value");
        return false;
      }

      open();

      createCategory(catVal);

      return true;
    } finally {
      close();
    }
  }

  private BwCategory createCategory(final String catVal) throws Throwable {
    if (pstate.debug) {
      trace("About to create category " + catVal);
    }

    BwCategory cat = BwCategory.makeCategory();

    cat.setWordVal(catVal);
    //cat.setOwner(svci.getUser());

    getSvci().getCategoriesHandler().add(cat);

    return cat;
  }

  private boolean createCollection() throws Throwable {
    try {
      open();

      String type = word();
      int calType;
      boolean topicalArea  = false;

      if (type == null) {
        error("Expected a collection type");
        return false;
      }

      if ("folder".equals(type)) {
        calType = BwCalendar.calTypeFolder;
      } else if ("calendar".equals(type)) {
        calType = BwCalendar.calTypeCalendarCollection;
      } else if ("alias".equals(type)) {
        calType = BwCalendar.calTypeAlias;
      } else if ("topic".equals(type)) {
        calType = BwCalendar.calTypeAlias;
        topicalArea  = true;
      } else {
        error("Expected a collection type 'folder', 'calendar', 'alias' or 'topic'");
        return false;
      }

      String parentPath = wordOrQuotedVal();

      if (parentPath == null) {
        if (pstate.debug) {
          trace("No parent path");
        }
        return false;
      }

      String calName = wordOrQuotedVal();

      if (calName == null) {
        error("Expected a collection name");
        return false;
      }

      BwCalendar cal = new BwCalendar();

      cal.setName(calName);
      cal.setSummary(calName);
      cal.setCalType(calType);
      cal.setPath(parentPath + "/" + calName);

      if (calType == BwCalendar.calTypeAlias) {
        BwCalendar target = getCal();

        if (target == null) {
          error("Require a target for alias");
          return false;
        }

        cal.setAliasUri("bwcal://" + target.getPath());

        if (topicalArea) {
          cal.setIsTopicalArea(true);
        }
      }

      /* filter on category="x" */

      String filterCatWd = null;

      if (pstate.tokenizer.testToken("filter")) {
        pstate.tokenizer.assertToken("on");

        pstate.tokenizer.assertToken("category");
        pstate.tokenizer.assertToken('=');

        filterCatWd = quotedVal();
      }

      List<String> cats = new ArrayList<String>();

      while (pstate.tokenizer.testToken("category")) {
        pstate.tokenizer.assertToken('=');
        cats.add(quotedVal());
      }

      String ownerAccount = word();
      String ownerHref;

      if (ownerAccount != null) {
        BwUser owner = getSvci().getUsersHandler().getUser(ownerAccount);

        if (owner == null) {
          error("No owner with account" + ownerAccount);
          return false;
        }

        ownerHref = owner.getPrincipalRef();
        cal.setOwnerHref(ownerHref);
        cal.setCreatorHref(ownerHref);
      } else {
        ownerHref = getSvci().getPrincipal().getPrincipalRef();
      }

      /* Now we have the owner find or add the categories */

      if (filterCatWd != null) {
        BwCategory cat = getCat(ownerHref, filterCatWd);

        if (cat == null) {
          error("Category " + filterCatWd + " not found for filter");
          return false;
        }

        cal.setFilterExpr("catuid=(\"" + cat.getUid() + "\")");
      }

      for (String catStr: cats) {
        BwCategory cat = getCat(ownerHref, catStr);

        if (cat != null) {
          cal.addCategory(cat);
        }
      }

      try {
        getSvci().getCalendarsHandler().add(cal, parentPath);
      } catch (CalFacadeException cfe) {
        if (CalFacadeException.duplicateCalendar.equals(cfe.getMessage())) {
          error("Collection " + calName + " already exists on path " + parentPath);
          return false;
        }

        if (CalFacadeException.collectionNotFound.equals(cfe.getMessage())) {
          error("Collection " + parentPath + " does not exist");
          return false;
        }

        throw cfe;
      }

      return true;
    } finally {
      close();
    }
  }

  private boolean createLocation() throws Throwable {
    try {
      open();

      BwLocation loc = BwLocation.makeLocation();

      if (!pstate.tokenizer.testToken("address")) {
        error("address required for location");
      }

      pstate.tokenizer.assertToken('=');

      loc.setAddress(new BwString(null, quotedVal()));

      if (pstate.tokenizer.testToken("subaddress")) {
        pstate.tokenizer.assertToken('=');

        loc.setSubaddress(new BwString(null, quotedVal()));
      }

      if (pstate.tokenizer.testToken("link")) {
        pstate.tokenizer.assertToken('=');

        loc.setLink(quotedVal());
      }

      getSvci().getLocationsHandler().add(loc);

      return true;
    } finally {
      close();
    }
  }
}
