/* ********************************************************************
    Licensed to Jasig under one or more contributor license
    agreements. See the NOTICE file distributed with this work
    for additional information regarding copyright ownership.
    Jasig licenses this file to you under the Apache License,
    Version 2.0 (the "License"); you may not use this file
    except in compliance with the License. You may obtain a
    copy of the License at:
        
    http://www.apache.org/licenses/LICENSE-2.0
        
    Unless required by applicable law or agreed to in writing,
    software distributed under the License is distributed on
    an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
    KIND, either express or implied. See the License for the
    specific language governing permissions and limitations
    under the License.
*/
package org.bedework.tools.cmdutil;

import org.bedework.calfacade.BwCalendar;
import org.bedework.calfacade.BwCategory;
import org.bedework.calfacade.BwEvent;
import org.bedework.calfacade.BwXproperty;
import org.bedework.calfacade.exc.CalFacadeDupNameException;
import org.bedework.calfacade.svc.EventInfo;
import org.bedework.tools.cmdutil.CmdUtil.ProcessState;

import edu.rpi.sss.util.Util;

import java.util.ArrayList;
import java.util.Collection;

/**
 * @author douglm
 *
 */
public class ProcessMove extends CmdUtilHelper {
  private int batchSize = 200;

  ProcessMove(final ProcessState pstate) {
    super(pstate);
  }

  @Override
  boolean process() throws Throwable {
    String wd = word();

    if (wd == null) {
      return false;
    }

    if ("events".equals(wd)) {
      return moveEvents();
    }

    return false;
  }

  private boolean moveEvents() throws Throwable {
    String toPath;
    String toOwner;
    String fromPath;

    /* Expect from (possibly quoted)
     *        to  (possibly quoted)
     *        0 or more categories
     */
    try {
      boolean addcats = pstate.tokenizer.testToken("addcats");

      boolean setname = pstate.tokenizer.testToken("setname");

      Collection<String> aliases = new ArrayList<String>();

      while (pstate.tokenizer.testToken("alias")) {
        pstate.tokenizer.assertToken('=');

        aliases.add(quotedVal());
      }

      open();

      BwCalendar from = getCal();

      if (from == null) {
        if (pstate.debug) {
          trace("No from cal");
        }
        return false;
      }

      fromPath = from.getPath();

      BwCalendar to = getCal();

      if (to == null) {
        if (pstate.debug) {
          trace("No to cal");
        }
        return false;
      }

      toOwner = to.getOwnerHref();
      toPath = to.getPath();

      close();

      /* Get the list of categories */

      Collection<String> catVals = new ArrayList<String>();

      while (!cmdEnd()) {
        String catVal = wordOrQuotedVal();

        if (catVal == null) {
          error("Expected a category");
          return false;
        }

        catVals.add(catVal);
      }

      /* Now we need to move the stuff in batches */
      for (;;) {
        open();

        Collection<BwCategory> cats = getCats(catVals,
                                              toOwner,
                                              addcats,
                                              fromPath);
        if (cats == null) {
          return false;
        }

        Collection<String> names = getSvci().getAdminHandler().getChildEntities(fromPath,
                                                                           0,
                                                                           batchSize);

        if (Util.isEmpty(names)) {
          break;
        }

        info("Move " + names.size() +
             " events from " + fromPath + " to " + toPath);
        long start = System.currentTimeMillis();

        for (String name: names) {
          moveEvent(fromPath,
                    toPath,
                    name,
                    cats,
                    aliases,
                    setname);
        }

        long elapsed = System.currentTimeMillis() - start;

        info("Move of " + names.size() +
             " events took " + elapsed / 1000 +
             "seconds at " + elapsed / names.size() +
             "millisecs per event");

        close();
      }

      return true;
    } finally {
      close();
    }
  }

  private void moveEvent(final String fromPath,
                         final String toPath,
                         final String name,
                         final Collection<BwCategory> cats,
                         final Collection<String> aliases,
                         final boolean setname) throws Throwable {
    EventInfo ei = getEvent(fromPath, name);

    BwEvent ev = ei.getEvent();

    if (pstate.debug) {
      StringBuilder sb = new StringBuilder("Moving event ");
      sb.append(ev.getUid());
      if (ev.getRecurring()) {
        sb.append(" recurring");
      }

      sb.append(" from ");
      sb.append(fromPath);
      sb.append(" to ");
      sb.append(toPath);

      trace(sb.toString());
    }

    ev.setColPath(toPath);

    for (BwCategory cat: cats) {
      ev.addCategory(cat);
    }

    for (String alias: aliases) {
      BwXproperty x = new BwXproperty();
      x.setName(BwXproperty.bedeworkAlias);
      x.setValue(alias);

      ev.addXproperty(x);
    }

    if (setname) {
      String nm = ev.getName();
      String nnm = ev.getUid() + ".ics";

      if (!nm.equals(nnm)) {
        ev.setName(nnm);
      }
    }

    if (ei.getOverrideProxies() != null) {
      for (BwEvent oev: ei.getOverrideProxies()) {
        oev.setColPath(toPath);
      }
    }

    try {
      getSvci().getEventsHandler().update(ei, false, null);
    } catch (CalFacadeDupNameException cdne) {
      pstate.errors.add("Duplicate name " + ev.getName() +
                        " uid: " + ev.getUid() +
                        " from: " + fromPath);
    }

  }

  private Collection<BwCategory> getCats(final Collection<String> catVals,
                                         final String toOwner,
                                         final boolean addcats,
                                         final String fromPath) throws Throwable {
    Collection<BwCategory> cats = new ArrayList<BwCategory>();

    for (String catVal: catVals) {
      BwCategory cat = getCat(toOwner, catVal);

      if (cat == null) {
        if (pstate.debug) {
          trace("No cat ");
        }
        return null;
      }

      cats.add(cat);
    }

    if (addcats) {
      String[] pathEls = fromPath.split("/");

      // First element should be "" for the leading "/"
      // Second element is the public root.

      if (pathEls.length > 2) {
        for (int i = 2; i < pathEls.length; i++) {
          String catVal = pathEls[i];
          BwCategory cat = getCat(toOwner, catVal);

          if (cat == null) {
            return null;
          }

          cats.add(cat);
        }
      }
    }

    return cats;
  }
}
