/* ********************************************************************
    Licensed to Jasig under one or more contributor license
    agreements. See the NOTICE file distributed with this work
    for additional information regarding copyright ownership.
    Jasig licenses this file to you under the Apache License,
    Version 2.0 (the "License"); you may not use this file
    except in compliance with the License. You may obtain a
    copy of the License at:
        
    http://www.apache.org/licenses/LICENSE-2.0
        
    Unless required by applicable law or agreed to in writing,
    software distributed under the License is distributed on
    an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
    KIND, either express or implied. See the License for the
    specific language governing permissions and limitations
    under the License.
*/
package org.bedework.tools.directory;

import edu.rpi.sss.util.Args;

import org.apache.commons.codec.binary.Base64;
import org.apache.log4j.Logger;

import java.security.MessageDigest;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import javax.naming.Context;
import javax.naming.directory.Attribute;
import javax.naming.directory.BasicAttribute;
import javax.naming.directory.DirContext;
import javax.naming.directory.ModificationItem;

/**
 * @author douglm
 *
 */
public class DirTool {
  private boolean debug = true;

  private transient Logger log;

  private String ldapUrl = "ldap://localhost:10389";
  private String baseDn = "dc=bedework, dc=org";

  private String accountsOu = "accounts";
  private String accountsDn = "ou=" + accountsOu + ", " + baseDn;
  private String accountsAttr = "uid";

  private String groupsOu = "groups";
  private String groupsDn = "ou=" + groupsOu + ", " + baseDn;

  private LdapDirectory ldir;

  private String adminId = "uid=admin,ou=system";

  private String adminPw = "secret";

  private static class UserInfo {
    String account;
    String firstName;
    String lastName;
    String caladdr;
    String password;

    GroupInfo group;
  }

  private static class GroupInfo extends UserInfo {
    List<UserInfo> members;

    void addMember(UserInfo val) {
      if (members == null) {
        members = new ArrayList<UserInfo>();
      }

      members.add(val);
    }
  }

  private boolean addUser;
  private UserInfo uinfo;

  private boolean userExists;
  String account;

  /* Just add root */
  private boolean init;

  private boolean cleanForQuickstart;

  private boolean addGroupMember;
  private boolean addGroup;
  private GroupInfo ginfo;

  private static final String pwEncryption = "SHA";

  /** Create the ous we need
   *
   */
  /**
   * @throws Throwable
   */
  private void cleanDirectory() throws Throwable {
    destroy(accountsDn);
    destroy(groupsDn);
  }

  private void destroy(String dn) throws Throwable {
    /* We cannot destroy a (sub)context that has any children */

    trace("Destroy " + dn);

    Directory.DirSearchResult sres = getLdir().searchOne(dn, null);

    if (sres != null) {
      for (;;) {
        DirRecord d = sres.nextRecord();

        if (d == null) {
          break;
        }

        destroy(d.getName() + "," + dn);
      }
    }

    getLdir().destroy(dn);
  }

  /** Create the ous we need
   *
   */
  /**
   * @return boolean true if created, false if already exists
   * @throws Throwable
   */
  private boolean addOus() throws Throwable {
    /** Build a directory record and add the attributes
     */
    DirRecord dirRec = new BasicDirRecord();

    String dn = accountsDn;
    dirRec.setDn(dn);
    dirRec.setAttr("ou", accountsOu);
    dirRec.setAttr("objectclass", "top");
    dirRec.setAttr("objectclass", "organizationalUnit");

    /* Ignore the result in case we got interrupted halfway. */
    getLdir().create(dirRec);

    dirRec = new BasicDirRecord();

    dn = groupsDn;
    dirRec.setDn(dn);
    dirRec.setAttr("ou", groupsOu);
    dirRec.setAttr("objectclass", "top");
    dirRec.setAttr("objectclass", "organizationalUnit");

    return getLdir().create(dirRec);
  }

  /** Create a user
   *
   * @param uinfo
   * @throws Throwable
   */
  private void createAccount(UserInfo uinfo) throws Throwable {
    /** Build a directory record and add the attributes
     */
    DirRecord dirRec = new BasicDirRecord();

    String userDn = accountsAttr + "=" + uinfo.account + ", " + accountsDn;
    dirRec.setDn(userDn);
    dirRec.setAttr(accountsAttr, uinfo.account);
    dirRec.setAttr("objectclass", "top");
    dirRec.setAttr("objectclass", "person");
    dirRec.setAttr("objectclass", "organizationalPerson");
    dirRec.setAttr("objectclass", "inetOrgPerson");

    String cn = uinfo.lastName + ", " + uinfo.firstName;
    dirRec.setAttr("cn", cn);
    //dirRec.setAttr("gecos", cn);
    dirRec.setAttr("sn", uinfo.lastName);
    dirRec.setAttr("mail", uinfo.caladdr);

    if (uinfo.password != null) {
      dirRec.setAttr("userPassword", encodedPassword(uinfo.password.toCharArray()));
    }

    /* Posix account requires these but we just set them to dummy values
     * /
    dirRec.setAttr("homeDirectory", "dummy");
    dirRec.setAttr("loginShell", "dummy");
    dirRec.setAttr("uidNumber", "999");
    dirRec.setAttr("gidNumber", "999");
    */

    getLdir().create(dirRec);
  }

  /** See if account exists
   *
   * @param account
   * @return boolean
   * @throws Throwable
   */
  private boolean accountExists(String account) throws Throwable {
    StringBuilder sb = new StringBuilder();

    sb.append("(&(");
    sb.append(accountsAttr);
    sb.append("=");
    sb.append(account);
    sb.append("))(objectClass=inetOrgPerson)");

    return getLdir().search(accountsDn,
                            sb.toString(),
                            Directory.scopeOne) != null;
  }

  /** Create a group - requires at least one member
   *
   * @param group
   * @throws Throwable
   */
  private void createGroup(GroupInfo group) throws Throwable {
    DirRecord dirRec = new BasicDirRecord();

    String dn = "cn=" + group.account + ", " + groupsDn;
    dirRec.setDn(dn);
    dirRec.setAttr("cn", group.account);
    dirRec.setAttr("objectclass", "top");
    dirRec.setAttr("objectclass", "groupOfNames");

    if (group.members != null) {
      for (UserInfo ui: group.members) {
        Attribute att = dirRec.getAttributes().get("member");

        if (att == null) {
          dirRec.setAttr("member", accountDn(ui.account));
        } else {
          att.add(accountDn(ui.account));
        }
      }
    }

    getLdir().create(dirRec);
  }

  /** Add user to a group
   *
   * @param group
   * @throws Throwable
   */
  private void addUserToGroup(GroupInfo group) throws Throwable {
    //if (!accountExists(account)) {
    //  error("Account " + account + " does not exist");
    //}

    BasicAttribute attr = new BasicAttribute("member",
                                             accountDn(group.members.get(0).account));
    ModificationItem mi = new ModificationItem(DirContext.ADD_ATTRIBUTE,
                                               attr);

    ModificationItem[] mods = {mi};
    getLdir().modify(groupDn(group.account), mods);
  }

  private String accountDn(String accountName) {
    return accountsAttr + "=" + accountName + ", " + accountsDn;
  }

  private String groupDn(String groupName) {
    return "cn=" + groupName + ", " + groupsDn;
  }

  private LdapDirectory getLdir() throws Throwable {
    if (ldir != null) {
      return ldir;
    }

    Properties pr = new Properties();

    pr.put(Context.PROVIDER_URL, ldapUrl);

    ldir = new LdapDirectory(pr, adminId, adminPw, debug);

    return ldir;
  }

  private String encodedPassword(char[] pw) throws Exception {
    MessageDigest md = MessageDigest.getInstance(pwEncryption);

    md.update(new String(pw).getBytes());

    byte[] b64s = new Base64().encode(md.digest());

    return "{" + pwEncryption + "}" + new String(b64s);
  }

  void process() throws Throwable {
    if (cleanForQuickstart) {
      if (debug) {
        trace("Trashing the directory");
      }
      cleanDirectory();
      if (!addOus()) {
        System.exit(1);
      }
    }

    if (init) {
      if (debug) {
        trace("Initialise directory");
      }
      if (!addOus()) {
        System.exit(1);
      }
    }

    if (addUser) {
      if (debug) {
        trace("Add user " + uinfo.account);
      }

      if (accountExists(uinfo.account)) {
        error("Account " + uinfo.account + " exists");
        return;
      }

      createAccount(uinfo);
    }

    if (userExists) {
      if (accountExists(account)) {
        System.exit(0);
      }

      System.exit(1);
    }

    if (addGroup) {
      if (debug) {
        trace("Add group " + ginfo.account);
      }

//      if (groupExists(accountPar)) {
//        error("Group " + groupPar + " exists");
//        return;
//      }

      createGroup(ginfo);
    }

    if (addGroupMember) {
      if (debug) {
        trace("Add group member " + ginfo.members.get(0).account +
              " to group " + ginfo.account);
      }

      addUserToGroup(ginfo);
    }
  }

  boolean processArgs(Args args) throws Throwable {
    if (args == null) {
      return true;
    }

    while (args.more()) {
      if (args.ifMatch("-debug")) {
        debug = true;
      } else if (args.ifMatch("-ndebug")) {
        debug = false;
      } else if (args.ifMatch("-url")) {
        ldapUrl = args.next();
      } else if (args.ifMatch("-addUser", 4)) {
        addUser = true;
        uinfo = getUserInfo(args);
      } else if (args.ifMatch("-userExists", 1)) {
        userExists = true;
        account = args.next();
      } else if (args.ifMatch("-addGroup", 2)) {
        addGroup = true;
        ginfo = getGroupAndMember(args);
      } else if (args.ifMatch("-addGroupMember", 2)) {
        addGroupMember = true;
        ginfo = getGroupAndMember(args);
      } else if (args.ifMatch("-init")) {
        init = true;
      } else if (args.ifMatch("-clean-for-quickstart")) {
        cleanForQuickstart = true;
      } else {
        error("Illegal argument: " + args.current());
        usage();
        return false;
      }
    }

    return true;
  }

  private GroupInfo getGroupAndMember(Args args) {
    GroupInfo gi = new GroupInfo();

    gi.account = args.next();
    UserInfo ui = new UserInfo();
    ui.account = args.next();
    gi.addMember(ui);

    return gi;
  }

  private UserInfo getUserInfo(Args args) {
    UserInfo ui = new UserInfo();
    ui.account = args.next();
    ui.firstName = args.next();
    ui.lastName = args.next();
    ui.caladdr = args.next();

    if (args.more() && !args.isMinusArg()) {
      ui.password = args.next();
    }

    return ui;
  }

  void usage() {
    System.out.println("Usage:");
    System.out.println("args   -debug");
    System.out.println("       -ndebug");
    System.out.println("       -url url");
    System.out.println("            set the ldap directory url");
    System.out.println("       -addUser account firstname lastname caladdr [password]");
    System.out.println("            create a new user account");
    System.out.println("       -userExists account");
    System.out.println("            check a user account exists");
    System.out.println("       -addGroup groupname groupmember");
    System.out.println("            Add a new group - first member is required");
    System.out.println("       -addGroupMember groupname groupmember");
    System.out.println("            Add a member to a group");
    System.out.println("       -init");
    System.out.println("            initialise the directory ");
    System.out.println("       -clean-for-quickstart");
    System.out.println("            clean the directory - deletes all accounts and groups ");
    System.out.println("            then creates new ous");
    System.out.println("");
  }

  protected Logger getLog() {
    if (log == null) {
      log = Logger.getLogger(this.getClass());
    }

    return log;
  }

  protected void error(String msg) {
    getLog().error(msg);
  }

  protected void trace(String msg) {
    getLog().debug(msg);
  }

  /** Main
   *
   * @param args
   */
  public static void main(String[] args) {
    DirTool dir = null;

    try {
      dir = new DirTool();

      if (!dir.processArgs(new Args(args))) {
        return;
      }

      dir.process();
    } catch (Throwable t) {
      t.printStackTrace();
    }
  }
}
